#include <stdlib.h>

#include "kernel.h"
#include "monitor.h"

/*
--------------------------------------------------------------------------------
CreateMonitor - create an initially unoccupied monitor
--------------------------------------------------------------------------------
*/

Monitor_t *
CreateMonitor(char *name)
{
	Monitor_t *mon = Malloc(sizeof(Monitor_t));

	mon->sem = CreateSem(name, 1);
	return mon;
}

/*
--------------------------------------------------------------------------------
DeleteMonitor - destroy a monitor
--------------------------------------------------------------------------------
*/

void 			
DeleteMonitor(Monitor_t *mon)
{
	DeleteSem(mon->sem);
	Free(mon);
}

/*
--------------------------------------------------------------------------------
EnterMonitor, EnterMonitorCond, EnterMonitorTimed - occupy a monitor

If return status is true, the task owns the monitor.
--------------------------------------------------------------------------------
*/

bool			
EnterMonitor(Monitor_t *mon)
{
	return EnterMonitorTimed(mon, FOREVER);
}

bool			
EnterMonitorCond(Monitor_t *mon)
{
	return EnterMonitorTimed(mon, 0);
}

bool			
EnterMonitorTimed(Monitor_t *mon, Time_t msecs)
{
	if ( mon->owner == mt_curr_task )
		Panic("EnterMonitorTimed: task already owns the monitor");

	if ( WaitSemTimed(mon->sem, msecs) )
	{
		mon->owner = mt_curr_task;
		return true;
	}
	return false;
}

/*
--------------------------------------------------------------------------------
LeaveMonitor - leave a monitor
--------------------------------------------------------------------------------
*/

void			
LeaveMonitor(Monitor_t *mon)
{
	if ( mon->owner != mt_curr_task )
		Panic("LeaveMonitor: task does not own the monitor");

	mon->owner = NULL;
	SignalSem(mon->sem);
}

/*
--------------------------------------------------------------------------------
CreateCondition - create a condition variable associated to a monitor
--------------------------------------------------------------------------------
*/

Condition_t *		
CreateCondition(char *name, Monitor_t *mon)
{
	Condition_t *cond = Malloc(sizeof(Condition_t));

	cond->queue = CreateQueue(name);
	cond->monitor = mon;
	return cond;
}

/*
--------------------------------------------------------------------------------
DeleteCondition - destroy a condition variable
--------------------------------------------------------------------------------
*/

void				
DeleteCondition(Condition_t *cond)
{
	DeleteQueue(cond->queue);
	Free(cond);
}

/*
--------------------------------------------------------------------------------
WaitCondition, WaitConditionCond, WaitConditionTimed - wait on a condition

The waiting task must own the monitor. If the task blocks, it atomically
leaves the monitor, waits in the task queue associated to the monitor and then
re-enters the monitor, returning the WaitQueue() status. 
This atomic condition guarantees that no wakeups are lost.
--------------------------------------------------------------------------------
*/

bool				
WaitCondition(Condition_t *cond)
{
	return WaitConditionTimed(cond, FOREVER);
}

bool				
WaitConditionTimed(Condition_t *cond, Time_t msecs)
{
	bool success;
	Monitor_t *mon = cond->monitor;

	if ( mon->owner != mt_curr_task )
		Panic("WaitConditionTimed: task does not own the monitor");

	Atomic();
	LeaveMonitor(mon);
	success = WaitQueueTimed(cond->queue, msecs);
	while ( !EnterMonitor(mon) )	// The monitor logic dictates that we MUST
		;							// re-enter the monitor here.
	Unatomic();

	return success;
}

/*
--------------------------------------------------------------------------------
SignalCondition - signal a condition variable.

The signalling task must own the monitor. This call wakes up the highest 
priority task waiting on the condition and returns true, or returns false if
the condition variable's queue task is empty. The awakened task completes its
WaitCondition() call successfully.
--------------------------------------------------------------------------------
*/

bool				
SignalCondition(Condition_t *cond)
{
	if ( cond->monitor->owner != mt_curr_task )
		Panic("SignalCondition: task does not own the monitor");

	return SignalQueue(cond->queue);
}

/*
--------------------------------------------------------------------------------
BroadcastCondition -  broadcast signalling of a condition variable

Same as SignalCondition, but wakes up all tasks waiting on the condition.
--------------------------------------------------------------------------------
*/

void				
BroadcastCondition(Condition_t *cond)
{
	if ( cond->monitor->owner != mt_curr_task )
		Panic("BroadcastCondition: task does not own the monitor");

	FlushQueue(cond->queue, true);
}
