#include <stdlib.h>

#include "mtask.h"
#include "kernel.h"

static TaskQueue_t time_q;

/*
--------------------------------------------------------------------------------
mt_enqueue - put a task to sleep in a task queue

The queue is ordered by increasing priority. If the two consecutive tasks have
the same priority, the second one has been sleeping for longer. The last task
in the queue has the highest priority and has been sleeping for the longest.
--------------------------------------------------------------------------------
*/

void 
mt_enqueue(Task_t *task, TaskQueue_t *queue)
{
	Task_t *ta;

	// Cheack where to insert
	for ( ta = queue->head ; ta && task->priority > ta->priority ;
															ta = ta->next )
		;
	if ( ta )						// insert before ta
	{
		if ( (task->prev = ta->prev) )
			ta->prev->next = task;
		else
			queue->head = task;
		ta->prev = task;
		task->next = ta;
	}
	else if ( (ta = queue->tail) )	// insert at the end
	{
		ta->next = queue->tail = task;
		task->prev = ta;
		task->next = NULL;
	}
	else							// the queue is empty
	{
		queue->head = queue->tail = task;
		task->next = task->prev = NULL;
	}
	task->queue = queue;
}

/*
--------------------------------------------------------------------------------
mt_dequeue - removes a task from a queue
--------------------------------------------------------------------------------
*/

void 
mt_dequeue(Task_t *task)
{
	TaskQueue_t *queue;

	if ( !(queue = task->queue) )
		return;
	if ( task->prev )
		task->prev->next = task->next;
	else
		queue->head = task->next;
	if ( task->next )
		task->next->prev = task->prev;
	else
		queue->tail = task->prev;
	task->next = task->prev = NULL;
	task->queue = NULL;
}

/*
--------------------------------------------------------------------------------
mt_peeklast, mt_getlast - access the last task in a queue

Mt_getlast() also removes the task from the queue.
--------------------------------------------------------------------------------
*/

Task_t *
mt_peeklast(TaskQueue_t *queue)
{
	return queue->tail;
}

Task_t *
mt_getlast(TaskQueue_t *queue)
{
	Task_t *task;

	if ( !(task = queue->tail) )
		return NULL;
	if ( (queue->tail = task->prev) )
		queue->tail->next = NULL;
	else
		queue->head = NULL;
	task->prev = task->next = NULL;
	task->queue = NULL;
	return task;
}

/*
--------------------------------------------------------------------------------
mt_enqueue_time - insert a task in the time queue

The time queue keeps tasks ordered by wakeup time. The msecs field in the
task control structure stores the difference in milliseconds between the 
wakeup time of each task and the previous one. For the first task in the queue,
the msecs field stores the absolute milliseconds remaining until wakeup.
The timer interrupt handler decrements this field and when it is less than the
time slice it removes the task from the time queue and makes it ready.
--------------------------------------------------------------------------------
*/

void 
mt_enqueue_time(Task_t *task, Time_t msecs)
{
	Task_t *ta;

	// Check where to insert
	for ( ta = time_q.head ; ta && msecs > ta->msecs ;
			msecs -= ta->msecs, ta = ta->time_next )
		;
	if ( ta )						// insert before ta
	{
		if ( (task->time_prev = ta->time_prev) )
			ta->time_prev->time_next = task;
		else
			time_q.head = task;
		ta->time_prev = task;
		task->time_next = ta;
		ta->msecs -= msecs;
	}
	else if ( (ta = time_q.tail) )	// insert at the end
	{
		ta->time_next = time_q.tail = task;
		task->time_prev = ta;
		task->time_next = NULL;
	}
	else							// the queue is empty
	{
		time_q.head = time_q.tail = task;
		task->time_next = task->time_prev = NULL;
	}
	task->msecs = msecs;
	task->in_time_q = true;
}

/*
--------------------------------------------------------------------------------
mt_dequeue_time - remove a task from the time queue
--------------------------------------------------------------------------------
*/

void 
mt_dequeue_time(Task_t *task)
{
	if ( !task->in_time_q )
		return;
	if ( task->time_prev )
		task->time_prev->time_next = task->time_next;
	else
		time_q.head = task->time_next;
	if ( task->time_next )
	{
		task->time_next->msecs += task->msecs;
		task->time_next->time_prev = task->time_prev;
	}
	else
		time_q.tail = task->time_prev;
	task->time_next = task->time_prev = NULL;
	task->in_time_q = false;
}

/*
--------------------------------------------------------------------------------
mt_peekfirst_time, mt_getfirst_time - access the first task in the time queue.

Mt_getfirst_time() also removes the task from the queue.
--------------------------------------------------------------------------------
*/

Task_t *
mt_peekfirst_time(void)
{
	return time_q.head;
}

Task_t *
mt_getfirst_time(void)
{
	Task_t *task;

	if ( !(task = time_q.head) )
		return NULL;
	if ( (time_q.head = task->time_next) )
	{
		time_q.head->time_prev = NULL;
		time_q.head->msecs += task->msecs;
	}
	else
		time_q.tail = NULL;
	task->time_prev = task->time_next = NULL;
	task->in_time_q = false;
	return task;
}
